/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 21 decembre 2006
   derniere modification le 08/03/2007 */


/**

IMPORTANT !! :
___
\p---> permettre de modifier la position de depart (& point de vue)
\p---> camera TPS. (mouvement avec point de fixation en rotation)

**/


#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEMath.h>
#include <SCE/utils/SCEVector.h>

#include <SCE/core/SCECMatrix.h>

#include <SCE/interface/SCECam.h>


void SCE_Cam_Init(SCE_SCamera *cam)
{
    if(cam != NULL)
    {
        /* mode par defaut: point de fixation */
        cam->mode = SCE_CAM_MODE_FIX;
        
        /* viewport par defaut */
        cam->viewport.x = cam->viewport.y = 0;
        cam->viewport.w = cam->viewport.h = 0; /* NOTE: dimensions de l'ecran */
        
        
        cam->p[0] = cam->p[1] = cam->p[2] = 0.0;
        cam->v[0] = cam->v[1] = cam->v[2] = 0.0;
        cam->i[0] = cam->i[1] = cam->i[2] = 0.0;
        
        /* position par defaut */
        cam->p[1] = 16.0;
        
        /* inclinaison: camera verticale par defaut */
        cam->i[2] = 1.0;
        
        /* point de vue par defaut: 0,0,0 */
        
        
        /* utilisation de l'inertie: non par defaut */
        cam->use_rot_inert = cam->use_mov_inert = 0;
        
        /* utilisation des deplacements additionnels, non par defaut */
        cam->use_additionnal_mov = cam->use_additionnal_rot = 0;
        
        
        /* distance par rapport au point de fixation par defaut */
        cam->fix_lenght = 16.0;
        
        
        /* angles de rotation par defaut : 0 */
        cam->relz = cam->relx = 0.0;
        cam->iz = cam->ix = 0.0;
        cam->z = cam->x = 0.0;
        
        
        /* inertie: aucune par defaut */
        cam->r_in = cam->m_in = 100.0;
        cam->r_pass = cam->m_pass = 0.0;
        cam->r_fois = cam->m_fois = 0;
        
        
        /* mouvements par defaut: aucuns */
        cam->front = cam->left = cam->f_rel = cam->l_rel = 0.0;
        cam->f_miss = cam->l_miss = 0.0;
        cam->l_fois = cam->f_fois = 0;
    }
}




void SCE_Cam_SetViewport(SCE_SCamera *cam, int x, int y, int w, int h)
{
    if(cam != NULL)
    {
        cam->viewport.x = x;
        cam->viewport.y = y;
        cam->viewport.w = w;
        cam->viewport.h = h;
    }
}

void SCE_Cam_SetMoveInert(SCE_SCamera *cam, float inert, int np)
{
    if(cam != NULL)
    {
        cam->m_in = inert;
        cam->m_pass = np;
    }
}

void SCE_Cam_SetRotateInert(SCE_SCamera *cam, float inert, int np)
{
    if(cam != NULL)
    {
        cam->r_in = inert;
        cam->r_pass = np;
    }
}

void SCE_Cam_SetMode(SCE_SCamera *cam, unsigned int mode)
{
    if(cam != NULL)
    {
        cam->mode = mode;
    }
}

void SCE_Cam_UseRotInert(SCE_SCamera *cam, int use)
{
    if(cam != NULL)
    {
        cam->use_rot_inert = use;
    }
}

void SCE_Cam_UseMovInert(SCE_SCamera *cam, int use)
{
    if(cam != NULL)
    {
        cam->use_mov_inert = use;
    }
}



void SCE_Cam_SetPointView(SCE_SCamera *cam, SCE_TVector3 view)
{
    if(cam != NULL)
    {
        switch(cam->mode)
        {
            case SCE_CAM_MODE_FREE:
                cam->v[0] = view[0] - cam->p[0];
                cam->v[1] = view[1] - cam->p[1];
                cam->v[2] = view[2] - cam->p[2];
                
                SCE_Vector3_Normalize(cam->v);
                
                break;
            
            case SCE_CAM_MODE_FIX:
                cam->v[0] = view[0];
                cam->v[1] = view[1];
                cam->v[2] = view[2];
                break;
            
            #ifdef __SCE_DEBUG__
            default:
                printf("erreur: mauvais mode de la camera\n");
            #endif
        }
    }
}


/* fonctions statiques */

/* calcul des angles de rotation avec l'inertie */
static void SCE_Cam_ComputeRotations(SCE_SCamera *cam)
{
    if(cam->x == 0 && cam->z == 0)
        cam->r_fois++;
    else
        cam->r_fois = 0;
    
    
    cam->relx += cam->x;
    cam->relz += cam->z;
    
    
    if(cam->r_fois >= cam->r_pass)
    {
        cam->ix = cam->relx;
        cam->iz = cam->relz;
    }
    else
    {
        cam->ix += (cam->relx - cam->ix) * (cam->r_in/100.0);
        cam->iz += (cam->relz - cam->iz) * (cam->r_in/100.0);
    }
}

/* calcul du mouvement avec l'inertie */
static void SCE_Cam_ComputeMovements(SCE_SCamera *cam)
{
    /* mouvement avant-arriere */
    if(cam->f_rel == 0.0)
        cam->f_fois++;
    else
        cam->f_fois = 0;
    
    if(cam->f_fois >= cam->m_pass)
    {
        cam->f_fois = cam->m_pass;
        cam->front = cam->f_rel+cam->f_miss;
    }
    else
    {
        cam->front = ((cam->f_rel+cam->f_miss)-cam->front) * (cam->m_in/100.0);
    }
    
    cam->f_miss += cam->f_rel - cam->front;
    /***/
    
    
    /* mouvement gauche-droite */
    if(cam->l_rel == 0.0)
        cam->l_fois++;
    else
        cam->l_fois = 0;
    
    if(cam->l_fois >= cam->m_pass)
    {
        cam->l_fois = cam->m_pass;
        cam->left = cam->l_rel+cam->l_miss;
    }
    else
    {
        cam->left = ((cam->l_rel+cam->l_miss)-cam->left) * (cam->m_in/100.0);
    }
    
    cam->l_miss += cam->l_rel - cam->left;
    /***/
}



void SCE_Cam_Rotate(SCE_SCamera *cam, float z, float x)
{
    if(cam != NULL)
    {
        cam->z += z;
        cam->x += x;
    }
}


void SCE_Cam_Move(SCE_SCamera *cam, float f, float l)
{
    if(cam != NULL)
    {
        cam->f_rel += f;
        cam->l_rel += l;
    }
}


void SCE_Cam_Compute(SCE_SCamera *cam)
{
    SCE_TVector3 vec;
    
    if(cam != NULL)
    {
        
        /* calcul de l'inertie si utilisee,
           sinon, simple calcul sans inertie */
        
        if(cam->use_rot_inert)
            SCE_Cam_ComputeRotations(cam);
        else
        {
            /* simples additions de rotation */
            cam->ix += cam->x; cam->iz += cam->z;
        }
        if(cam->use_mov_inert)
            SCE_Cam_ComputeMovements(cam);
        else
        {
            /* simples effectations de mouvement */
            cam->front = cam->f_rel; cam->left = cam->l_rel;
        }
        
        switch(cam->mode)
        {
            /* mode 'fps' */
            case SCE_CAM_MODE_FREE:
                cam->v[0] = cam->v[2] = 0.0;
                cam->v[1] = 1.0;
                
                /* rotations */
                SCE_Vector3_RotateX(cam->v, cos(cam->ix*RAD), sin(cam->ix*RAD));
                SCE_Vector3_RotateZ(cam->v, cos(cam->iz*RAD), sin(cam->iz*RAD));
                
                /* deplacement frontal */
                cam->p[0] += cam->v[0] * cam->front;
                cam->p[1] += cam->v[1] * cam->front;
                cam->p[2] += cam->v[2] * cam->front;
                
                /* deplacement latteral */
                /* construction du vecteur latteral */
                vec[0] = -cam->v[1];
                vec[1] = cam->v[0];
                vec[2] = cam->i[0];
                
                SCE_Vector3_Normalize(vec);
                
                cam->p[0] += vec[0] * cam->left;
                cam->p[1] += vec[1] * cam->left;
                cam->p[2] += vec[2] * cam->left;
                
                /* point de vue final */
                cam->v[0] += cam->p[0];
                cam->v[1] += cam->p[1];
                cam->v[2] += cam->p[2];
                
                break;
            
            /* mode 'visualisation' */
            case SCE_CAM_MODE_FIX:
                cam->p[0] = cam->v[0];
                
                /* calcul de la distance entre la camera
                   et le point de visualisation */
                cam->fix_lenght -=cam->front;
                cam->p[1] = cam->v[1] + cam->fix_lenght;
                
                cam->p[2] = cam->v[2];
                
                SCE_Vector3_RotateX(cam->p, -cos(cam->ix*RAD), -sin(cam->ix*RAD));
                SCE_Vector3_RotateZ(cam->p, cos(cam->iz*RAD), sin(cam->iz*RAD));
                
                break;
            
            #ifdef __SCE_DEBUG__
            default:
                printf("erreur: mauvais mode de la camera\n");
            #endif
        }
        
        
        if(!cam->use_additionnal_mov)
        {
            /* la camera n'utilise pas les mouvements additionnels,
               replacage des valeurs de mouvements a zero */
            cam->f_rel = 0.0;
            cam->l_rel = 0.0;
        }
        
        
        if(!cam->use_additionnal_rot)
        {
            /* la camera n'utilise pas les rotations additionnels,
               replacage des valeurs de rotations a zero */
            cam->x = 0.0;
            cam->z = 0.0;
        }
    }
}


void SCE_Cam_SetActive(const SCE_SCamera* const cam)
{
    if(cam != NULL)
    {
        SCE_CViewport(cam->viewport.x, cam->viewport.y,
                      cam->viewport.w, cam->viewport.h);
        
        SCE_CLookv(cam->p, cam->v, cam->i);
    }
}
