/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 3 novembre 2006
   derniere modification le 23/02/2007 */


#include <SCE/SCEMinimal.h>

#include <SCE/interface/SCEM3DLoader.h>


void* SCE_M3D_Load(FILE *file, const char *filename, void *n_objs)
{
    SCE_M3D_Model *obj = NULL;
    unsigned int i=0;
    int magic;
    unsigned int *nb_objs = NULL;
    unsigned int unsused; /* variable pointee par nb_objs si n_objs = NULL */
    
    
    /* utilite du nom du fichier ? aucune... */
    filename = NULL;
    
    /* verification de l'argument */
    if(n_objs)
        nb_objs = n_objs;
    else
        nb_objs = &unsused;
    
    
    fread(&magic, sizeof(int), 1, file);
    if(magic != SCE_M3D_MAGIC)
    {
        Logger_Log(SCE_BAD_FORMAT);
        return (NULL);
    }
    
    fread(nb_objs, sizeof(unsigned int), 1, file);
    if((*nb_objs) <= 0)
    {
        Logger_Log(SCE_BAD_FORMAT);
        return (NULL);
    }
    /***/
    
    
    /* allocation */
    obj = SCE_malloc(sizeof(*obj)*(*nb_objs));
    if(obj == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    /***/
    
    
    for(i=0; i<(*nb_objs); i++)
    {
        obj[i].material = NULL;
        obj[i].material = SCE_calloc(sizeof(*obj[i].material), 1);
        if(obj[i].material == NULL)
        {
            SCE_M3D_Free(obj, &i);
            Logger_LogSrc();
            return (NULL);
        }
        
        /* donnes vectorielles */
        fread(&obj[i].vsize, sizeof(unsigned int), 1, file);
        fread(&obj[i].nsize, sizeof(unsigned int), 1, file);
        fread(&obj[i].tsize, sizeof(unsigned int), 1, file);
        
        obj[i].vertices = SCE_calloc(sizeof(float), obj[i].vsize);
        if(obj[i].vertices == NULL && obj[i].vsize > 0)
        {
            SCE_M3D_Free(obj, &i);
            Logger_LogSrc();
            return (NULL);
        }
        obj[i].normals = SCE_calloc(sizeof(float), obj[i].nsize);
        if(obj[i].normals == NULL && obj[i].nsize > 0)
        {
            SCE_M3D_Free(obj, &i);
            Logger_LogSrc();
            return (NULL);
        }
        obj[i].texcoords = SCE_calloc(sizeof(float), obj[i].tsize);
        if(obj[i].texcoords == NULL && obj[i].tsize > 0)
        {
            SCE_M3D_Free(obj, &i);
            Logger_LogSrc();
            return (NULL);
        }
        
        fread(&obj[i].vertices[0], sizeof(float), obj[i].vsize, file);
        fread(&obj[i].normals[0], sizeof(float), obj[i].nsize, file);
        fread(&obj[i].texcoords[0], sizeof(float), obj[i].tsize, file);
        /***/
        
        
        /* indices */
        fread(&obj[i].lines_count, sizeof(unsigned int), 1, file);
        fread(&obj[i].triangles_count, sizeof(unsigned int), 1, file);
        fread(&obj[i].quads_count, sizeof(unsigned int), 1, file);
        
        if((obj[i].vsize/3) < 65536)
            obj[i].lines_index = SCE_calloc(sizeof(unsigned short), obj[i].lines_count);
        else
            obj[i].lines_index = SCE_calloc(sizeof(unsigned int), obj[i].lines_count);
        
        if(obj[i].lines_index == NULL && obj[i].lines_count > 0)
        {
            SCE_M3D_Free(obj, &i);
            Logger_LogSrc();
            return (NULL);
        }
        
        if((obj[i].vsize/3) < 65536)
            obj[i].triangles_index = SCE_calloc(sizeof(unsigned short), obj[i].triangles_count);
        else
            obj[i].triangles_index = SCE_calloc(sizeof(unsigned int), obj[i].triangles_count);
        
        if(obj[i].triangles_index == NULL && obj[i].triangles_count > 0)
        {
            SCE_M3D_Free(obj, &i);
            Logger_LogSrc();
            return (NULL);
        }
        
        if((obj[i].vsize/3) < 65536)
            obj[i].quads_index = SCE_calloc(sizeof(unsigned short), obj[i].quads_count);
        else
            obj[i].quads_index = SCE_calloc(sizeof(unsigned int), obj[i].quads_count);
        
        if(obj[i].quads_index == NULL && obj[i].quads_count > 0)
        {
            SCE_M3D_Free(obj, &i);
            Logger_LogSrc();
            return (NULL);
        }
        
        if((obj[i].vsize/3) < 65536)
        {
            fread((unsigned short*)obj[i].lines_index, sizeof(unsigned short), obj[i].lines_count, file);
            fread((unsigned short*)obj[i].triangles_index, sizeof(unsigned short), obj[i].triangles_count, file);
            fread((unsigned short*)obj[i].quads_index, sizeof(unsigned short), obj[i].quads_count, file);
        }
        else
        {
            fread((unsigned int*)obj[i].lines_index, sizeof(unsigned int), obj[i].lines_count, file);
            fread((unsigned int*)obj[i].triangles_index, sizeof(unsigned int), obj[i].triangles_count, file);
            fread((unsigned int*)obj[i].quads_index, sizeof(unsigned int), obj[i].quads_count, file);
        }
        /***/
        
        
        /* materiau */
        fread(&obj[i].material->ambient[0], sizeof(float), 4, file);
        fread(&obj[i].material->diffuse[0], sizeof(float), 4, file);
        fread(&obj[i].material->specular[0], sizeof(float), 4, file);
        fread(&obj[i].material->shininess, sizeof(float), 1, file);
        /***/
        
        if((obj[i].vsize/3) < 65536)
        {
            obj[i].type = GL_UNSIGNED_SHORT;
            obj[i].size = sizeof(unsigned short);
        }
        else
        {
            obj[i].type = GL_UNSIGNED_INT;
            obj[i].size = sizeof(unsigned int);
        }
    }
    /***/
    
    /* fini :) */
    return ((SCE_M3D_Model*)obj);
}


int SCE_M3D_Free(void *objs, void *n_objs)
{
    unsigned int i;
    unsigned int nb_objs;
    SCE_M3D_Model *model = NULL;
    
    if(n_objs)
        nb_objs = *(unsigned int*)n_objs;
    else
        nb_objs = 0;
    
    if(objs)
    {
        model = objs;
        
        for(i=0; i<nb_objs; i++)
        {
            if(SCE_free(model[i].vertices) < 0
             ||SCE_free(model[i].normals) < 0
             ||SCE_free(model[i].texcoords) < 0
             ||SCE_free(model[i].lines_index) < 0
             ||SCE_free(model[i].triangles_index) < 0
             ||SCE_free(model[i].quads_index) < 0
             ||SCE_free(model[i].material) < 0)
            {
                Logger_LogSrc();
                return (-1);
            }
        }
        
        if(SCE_free(model) < 0)
        {
            Logger_LogSrc();
            return (-1);
        }
    }
    
    return (0);
}


/* NOTE: debug mode only ...
void M3D_PrintObjs(const M3D_Obj* obj, const unsigned int nb_objs)
{
    unsigned int i=0, j=0;
    
    if(obj == NULL)
        return;
    
    for(j=0; j<nb_objs; j++)
    {
        printf("vertices : %d\nnormals : %d\ntexcoords : %d\n", obj->vsize/3, obj->nsize/3, obj->tsize/2);
        
        for(i=0; i<obj->vsize; i+=3)
            printf("vertex[%d %d %d] = %f %f %f\n", i, i+1, i+2, obj->vertices[i], obj->vertices[i+1], obj->vertices[i+2]);
        
        for(i=0; i<obj->nsize; i+=3)
            printf("normal[%d %d %d] = %f %f %f\n", i, i+1, i+2, obj->normals[i], obj->normals[i+1], obj->normals[i+2]);
        
        for(i=0; i<obj->tsize; i+=3)
            printf("texcoord[%d %d %d] = %f %f %f\n", i, i+1, i+2, obj->texcoords[i], obj->texcoords[i+1], obj->texcoords[i+2]);
        
        for(i=0; i<obj->triangles_count; i++)
            printf("triangles_index[%d] = %d\n", i, *((unsigned short*)obj->triangles_index+i));
        
        for(i=0; i<obj->quads_count; i++)
            printf("quads_index[%d] = %hu\n", i, *((unsigned short*)obj->quads_index+i));
    }
}*/
