/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 22 decembre 2006
   derniere modification le 01/05/2007 */


#include <stdlib.h>
#include <string.h>

#include <SCE/utils/SCEError.h>

#include <SCE/utils/SCEMemory.h>


typedef struct
{
    size_t size;
    
    unsigned int line;
    char *file;
} SCE_MemInfo;


static void** ptrs = NULL;
static size_t p_size = 0;


static int SCE_incr(void *ptr)
{
    p_size++;
    ptrs = realloc(ptrs, p_size*sizeof(void*));
    if(ptrs == NULL)
    {
        Logger_Log(SCE_OUT_OF_MEMORY);
        return (-1);
    }
    
    ptrs[p_size-1] = ptr; /* ;) */
    
    return (0);
}

static int SCE_decr(void *ptr)
{
    size_t i=0, j=0;
    
    while(i<p_size && ptrs[i] != ptr)
        i++;
    
    for(j=i; j<p_size-1; j++)
        ptrs[j] = ptrs[j+1];
    
    p_size--;
    
    ptrs = realloc(ptrs, p_size*sizeof(void*));
    if(p_size <= 0)
    {
        p_size = 0;
        ptrs = NULL;
    }
    else if(ptrs == NULL)
    {
        Logger_Log(SCE_OUT_OF_MEMORY);
        return (-1);
    }
    
    return (0);
}

static int SCE_verif(void *ptr)
{
    size_t i;
    
    for(i=0; i<p_size; i++)
        if(ptrs[i] == ptr)
            return (i);
    
    return (-1);
}

void* SCE_malloc_(size_t size, const char *file, const unsigned int line)
{
    SCE_MemInfo *meminfo = NULL;
    
    /* pointeur temporaire pour effectuer le transfert d'adresse */
    SCE_MemInfo **ptrans = NULL;
    
    void *ptr = NULL;
    
    /* allocation de notre structure */
    meminfo = malloc(sizeof *meminfo);
    if(meminfo == NULL)
    {
        Logger_Log(SCE_OUT_OF_MEMORY);
        return (NULL);
    }
    
    meminfo->size = size;
    meminfo->line = line;
    meminfo->file = NULL;
    meminfo->file = calloc(sizeof(char), strlen(file)+1);
    if(meminfo->file == NULL)
    {
        free(meminfo);
        if(strlen(file) == 0)
            Logger_Log(SCE_INVALID_ARG);
        else
            Logger_Log(SCE_OUT_OF_MEMORY);
        
        return (NULL);
    }
    strcpy(meminfo->file, file);
    
    ptr = malloc(size + sizeof(meminfo));
    if(ptr == NULL)
    {
        free(meminfo->file);
        free(meminfo);
        Logger_Log(SCE_OUT_OF_MEMORY);
        return (NULL);
    }
    
    ptrans = ptr;
    *ptrans = meminfo;
    
    if(SCE_incr(ptr) < 0)
    {
        free(meminfo->file);
        free(meminfo);
        free(ptr);
        
        Logger_LogSrc();
        
        return (NULL);
    }
    
    return (ptr+sizeof(meminfo));
}

void* SCE_calloc_(size_t size, size_t n_size,
                  const char *file, const unsigned int line)
{
    SCE_MemInfo *meminfo = NULL;
    
    /* pointeur temporaire pour effectuer le transfert d'adresse */
    SCE_MemInfo **ptrans = NULL;
    
    void *ptr = NULL;
    unsigned int i;
    
    /* allocation de notre structure */
    meminfo = malloc(sizeof *meminfo);
    if(meminfo == NULL)
    {
        Logger_Log(SCE_OUT_OF_MEMORY);
        return (NULL);
    }
    
    meminfo->size = size;
    meminfo->line = line;
    meminfo->file = NULL;
    meminfo->file = calloc(sizeof(char), strlen(file)+1);
    if(meminfo->file == NULL)
    {
        free(meminfo);
        if(strlen(file) <= 0)
            Logger_Log(SCE_INVALID_ARG);
        else
            Logger_Log(SCE_OUT_OF_MEMORY);
        
        return (NULL);
    }
    strcpy(meminfo->file, file);
    
    ptr = malloc((size*n_size) + sizeof(meminfo));
    if(ptr == NULL)
    {
        free(meminfo->file);
        free(meminfo);
        Logger_Log(SCE_OUT_OF_MEMORY);
        return (NULL);
    }
    
    for(i=0; i<size*n_size; i++)
        *((char*)(ptr+i)) = 0;
    
    ptrans = ptr;
    *ptrans = meminfo;
    
    if(SCE_incr(ptr) < 0)
    {
        free(meminfo->file);
        free(meminfo);
        free(ptr);
        
        Logger_LogSrc();
        
        return (NULL);
    }
    
    return (ptr+sizeof(meminfo));
}

void*
SCE_realloc_(void* aloc, size_t size, const char *file, const unsigned int line)
{
    /* pointeur temporaire pour effectuer le transfert d'adresse */
    SCE_MemInfo **ptrans = NULL;
    
    void *ptr = NULL;
    
    
    if(aloc != NULL) /* simple reallocation */
    {
        if(SCE_verif(aloc-sizeof(SCE_MemInfo*)) > -1)
        {
            ptr = aloc-sizeof(SCE_MemInfo*);
            
            if(size > 0)
            {
                ptr = realloc(ptr, size+sizeof(SCE_MemInfo*));
                if(ptr == NULL)
                {
                    Logger_Log(SCE_OUT_OF_MEMORY);
                    return (NULL);
                }
                
                ptrans = ptr;
                
                ptrs[SCE_verif(aloc-sizeof(SCE_MemInfo*))] = ptr;
                
                (*ptrans)->size = size;
                (*ptrans)->line = line;
                free((*ptrans)->file);
                (*ptrans)->file = NULL;
                (*ptrans)->file = malloc(strlen(file)*sizeof(char));
                if(file == NULL)
                {
                    Logger_Log(SCE_OUT_OF_MEMORY);
                    SCE_free(ptr);
                    return (NULL);
                }
                
                return (ptr+sizeof(SCE_MemInfo*));
            }
            else /* si taille 0 demandee --> liberation */
            {
                if(SCE_free(aloc) < 0)
                    Logger_LogSrc();
                
                ptr = NULL;
            }
        }
        else /* pointeur invalide */
        {
            Logger_Log(SCE_INVALID_POINTER);
            ptr = NULL;
        }
    }
    else /* allocation : creation d'un meminfo */
    {
        ptr = SCE_malloc_(size, file, line);
    }
    
    return (ptr);
}

int SCE_free_(void* aloc)
{
    /* pointeur temporaire pour effectuer le transfert d'adresse */
    SCE_MemInfo **ptrans = NULL;
    
    if(aloc != NULL)
    {
        if(SCE_verif(aloc-sizeof(SCE_MemInfo*)) > -1)
        {
            ptrans = aloc-sizeof(SCE_MemInfo*);
            
            free((*ptrans)->file);
            free(*ptrans);
            free(ptrans);
            
            if(SCE_decr(ptrans) < 0)
            {
                Logger_LogSrc();
                return (-1);
            }
        }
        else
        {
            Logger_Log(SCE_INVALID_POINTER);
            return (-1);
        }
    }
    
    return (0);
}

size_t SCE_Memory_GetSize(void* ptr)
{
    /* pointeur temporaire pour effectuer le transfert d'adresse */
    SCE_MemInfo **ptrans = NULL;
    
    ptrans = ptr-sizeof(SCE_MemInfo*);
    
    if(SCE_verif(ptr-sizeof(SCE_MemInfo*)) > -1)
        return ((*ptrans)->size);
    
    return (0);
}

char* SCE_Memory_GetFile(void* ptr)
{
    /* pointeur temporaire pour effectuer le transfert d'adresse */
    SCE_MemInfo **ptrans = NULL;
    
    ptrans = ptr-sizeof(SCE_MemInfo*);
    
    if(SCE_verif(ptr-sizeof(SCE_MemInfo*)) > -1)
        return ((*ptrans)->file);
    
    return (NULL);
}

unsigned int SCE_Memory_GetLine(void* ptr)
{
    /* pointeur temporaire pour effectuer le transfert d'adresse */
    SCE_MemInfo **ptrans = NULL;
    
    ptrans = ptr-sizeof(SCE_MemInfo*);
    
    if(SCE_verif(ptr-sizeof(SCE_MemInfo*)) > -1)
        return ((*ptrans)->line);
    
    return (0);
}
